# -*- coding: utf-8 -*-

from .config import permissions, validate_ip
from .exceptions import (
    CyberarkAPIException, CyberarkException, UserAlreadyMember, UserNotExists, SafeAlreadyExistsError
)
import json
import re
from .abstract import Bastion


class PrivilegedAccount:
    """Base class to be used with accounts fonctions"""
    def __init__(self, name: str, address: str, username: str, platformId: str, safeName: str,
                 platformAccountProperties: dict, secret: str = "", secretType:str = "Key",
                 secretManagement: dict = None,
                 remoteMachinesAccess: dict = None,
                 id: str = "",
                 **other):
        self.secret = secret
        if remoteMachinesAccess is None:
            remoteMachinesAccess = {"remoteMachines": "string", "accessRestrictedToRemoteMachines": True}
        if secretManagement is None:
            secretManagement = {"automaticManagementEnabled": True, "manualManagementReason": ""}
        self.remoteMachinesAccess = remoteMachinesAccess
        self.secretManagement = secretManagement
        self.secretType = secretType
        self.platformAccountProperties = platformAccountProperties
        self.safeName = safeName
        self.platformId = platformId
        self.username = username
        self.address = address
        self.name = name
        self.id = id
         
    def get_name(self):
        return "{}-{}".format(self.address, self.username)

    def to_json(self):
        json_object = {"id": self.id, "name": self.get_name(), "address": self.address, "userName": self.username,
                       "platformId": self.platformId, "safeName": self.safeName, "secret": self.secret,
                       "platformAccountProperties": self.platformAccountProperties,
                       "secretManagement": self.secretManagement}

        return json_object

    def __str__(self):
        strrepr = self.to_json()
        strrepr["secret"] = "***********"
        return str(strrepr)


class Account:
    def __init__(self, epv: Bastion):
        self.epv = epv

    def add_account_to_safe(self, account: PrivilegedAccount):
        url, head = self.epv.get_url('API/Accounts')

        try:
            req = self.epv.session.post(url, headers=head, data=json.dumps(account.to_json()), **self.epv.reqparms)
        except Exception as err:
            # uncatched exception
            raise CyberarkException(err)

        if req.status_code != 201:
            content = json.loads(req.content)
            raise CyberarkAPIException(content["ErrorMessage"])

        acc_id = json.loads(req.content)['id']
        account.id = acc_id


    def is_valid_username(name: str) -> bool:
        return re.match("[^@]+@[^@]+[.][^@]+", name) is not None


    def is_valid_safename(name: str) -> bool:
        return re.match("^[0-9]{12}$", name) is not None


    def user_exists(self, username: str):
        if self.epv.user_list is None:
            url, head = self.epv.get_url('API/Users')

            req = self.epv.session.get(url, headers=head, **self.epv.reqparms)
            if req.status_code != 200:
                raise Exception("Error occured during user_exists call. CyberArk error:", req.content)
            self.epv.user_list = [u['username'].lower().strip() for u in json.loads(req.content)['Users']]
        return username in self.epv.user_list


    def add_safe_member(self, safe: str, username: str, profile: str):
        url, head = self.epv.get_url('WebServices/PIMServices.svc/Safes/' + safe + '/Members')

        data = {'member': {
            'MemberName': username,
            'Permissions': permissions(profile)
        }
        }
        req = self.epv.session.post(url, headers=head, data=json.dumps(data), **self.epv.reqparms)
        if req.status_code != 201:
            content = json.loads(req.content)
            if "Reason: User or Group is already a member." in content["ErrorMessage"]:
                raise UserAlreadyMember
            elif "Reason: Input parameter for [MemberName] value is invalid" in content["ErrorMessage"]:
                raise UserNotExists
            else:
                raise Exception("Error occured during add_safe_member call. CyberArk error:", req.content)

    def remove_safe_member(self, safe: str, username: str):
        url, head = self.epv.get_url('WebServices/PIMServices.svc/Safes/' + safe + '/Members/' + username)
        req = self.epv.session.delete(url, headers=head, **self.epv.reqparms)

        if req.status_code != 200:
            raise Exception("Error occured during remove_safe_member call. CyberArk error:", req.content)

    def safe_exists(self, safename: str):
        url, head = self.epv.get_url('WebServices/PIMServices.svc/Safes/?query=' + safename)
        req = self.epv.session.get(url, headers=head, **self.epv.reqparms)
        if req.status_code == 200:
            return True
        elif req.status_code == 404:
            return False
        else:
            raise Exception("Error occured during remove_safe_member call. CyberArk error:", req.content)

    def create_safe(self, safename: str, description=""):
        url, head = self.epv.get_url('WebServices/PIMServices.svc/Safes')
        data = {
            "safe": {
                "SafeName": safename,
                "Description": description,
                "OLACEnabled": False,
                "ManagingCPM": self.epv.cpm,
                "NumberOfVersionsRetention": self.epv.retention
            }
        }

        try:
            req = self.epv.session.post(url=url, data=json.dumps(data), headers=head, **self.epv.reqparms)
        except Exception:
            raise CyberarkException

        if req.status_code != 201:
            error_code = req.json()["ErrorCode"]
            if error_code == 'CAWS00001E':
                raise SafeAlreadyExistsError()
            else:
                raise CyberarkException("Error occured during add_safe call. CyberArk error:")
        return 0


    def add_safe(self, safename: str, description=""):
        try:
            self.create_safe(safename, description)
        except Exception as e:
            raise e

        # Define Safe defaults owners
        for user, profile in {"Vault Admins": "admin", "onboarding": "prov", "Administrator": 'admin'}.items():
            try:
                self.add_safe_member(safename, user, profile)
            except UserAlreadyMember:
                pass
        # when we did that, we had a bug so we remove it
        # try:
        #     # Removing provisioning user from admin of the safe
        #     self.remove_safe_member(safename, config.username)
        # except CyberarkException:
        #     raise CyberarkException("Unable to remove %s from safe", config.username)


    def search_account_by_ip_addr(self, account: (PrivilegedAccount, str)) -> list:
        if type(account) is str:
            address = account
        elif isinstance(account, PrivilegedAccount):
            address = account.address
        else:
            raise TypeError("search account function returned : First argument of account is not valid")

        if not validate_ip(address):
            raise TypeError("The address of the object was not considered as valid IPv4 address %s", address)

        url, head = self.epv.get_url('API/Accounts?search=' + address)

        if isinstance(account, PrivilegedAccount):
            # Searching for both username + IP Address
            url += " " + account.username + "&limit=111"
        else:
            url += "&limit=111"

        req = self.epv.session.get(url=url, headers=head, verify=False)

        if req.status_code != 200:
            raise Exception("Error occured during search_account call. CyberArk error:")

        json_response_data = json.loads(req.content)
        exact_matches = []
        for result in json_response_data['value']:
            if result['platformAccountProperties']['IP'] == address:
                exact_matches.append(result)
        return exact_matches

    def search_account(self, expression: str) -> list:
        url, head = self.epv.get_url("API/Accounts?search=" + expression)
        req = self.epv.session.get(url=url, headers=head, verify=False)
        if req.status_code != 200:
            raise CyberarkException("An error occured during search_account call.")

        return json.loads(req.content)['value']